<?php

namespace Drupal\toolshed_media\Plugin\Field\FieldFormatter;

use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\Plugin\Field\FieldFormatter\EntityReferenceFormatterBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\file\Entity\File;
use Drupal\toolshed_media\Utility\FileHelper;
use Drupal\toolshed_media\Utility\MimeHelperInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A Field formatter for displaying file information for file or media entities.
 *
 * @FieldFormatter(
 *   id = "toolshed_file_info",
 *   label = @Translation("File Info Format"),
 *   field_types = {
 *     "file",
 *     "entity_reference",
 *   },
 * )
 */
class FileInfoFormatter extends EntityReferenceFormatterBase {

  // Entity types that can be used with the file info formatter works with.
  const ALLOWED_ENTITY_TYPES = ['file', 'media'];

  /**
   * Mime helper to transform extensions to friendly display types.
   *
   * @var \Drupal\toolshed_media\Utility\MimeHelperInterface
   */
  protected $mimeManager;

  /**
   * Creates a new instance of the FileInfoFormatter formatter plugin.
   *
   * @param array $configuration
   *   Array of plugin configuration options which includes the field formatter
   *   definition information:
   *     - $field_definition
   *     - $settings
   *     - $label
   *     - $view_mode
   *     - $third_party_settings
   *   These are the usual values for constructing a standard field formatter.
   * @param string $plugin_id
   *   The plugin ID.
   * @param mixed $plugin_definition
   *   The plugin definition.
   * @param \Drupal\toolshed_media\Utility\MimeHelperInterface $mime_helper
   *   Toolshed media mime manager. Maintains a map of file extensions to
   *   configured mime display settings.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, MimeHelperInterface $mime_helper) {
    parent::__construct(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings']
    );

    $this->mimeManager = $mime_helper;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('toolshed.media.mime_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $fieldDef) {
    $storageDef = $fieldDef->getFieldStorageDefinition();

    switch ($storageDef->getType()) {
      case 'entity_reference':
        $fieldType = $storageDef->getSetting('target_type');
        return in_array($fieldType, static::ALLOWED_ENTITY_TYPES);

      case 'file':
        return TRUE;
    }

    return FALSE;
  }

  /**
   * Get the list of available options for linking the entity reference field.
   *
   * For media it is valid for the resulting link to point to either the file
   * or the original media object. For file, the entity is the file, so they
   * actually both generate the same link.
   *
   * @return array
   *   Return a key/value list of available options for linking the field.
   */
  public function getLinkOptions() {
    return [
      'file' => $this->t('File'),
      'entity' => $this->t('Referenced entity'),
    ];
  }

  /**
   * Convert the number of raw bytes into a human friendly readable format.
   *
   * This will convert the bytes integer to a string of the file size with
   * the appropriate unit suffix.
   *
   * @param int $bytes
   *   The size of a file in number of bytes.
   * @param int $decimals
   *   Number of decimal places to keep in the returned string.
   *
   * @return string
   *   Return the file size with a unit suffix, and up to the number
   *   of decimal places requested.
   */
  protected function readableFileSize($bytes, $decimals = 2) {
    $i = 0;
    $divisor = 1;
    $next = 1024;
    $suffixes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'];

    while (($bytes / $next) > 1.0 && isset($suffixes[$i])) {
      $divisor = $next;
      $next *= 1024;
      ++$i;
    }

    return number_format($bytes / $divisor, $decimals) . ' ' . $suffixes[$i];
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $entityType = $this->fieldDefinition->getSetting('target_type');

    if (in_array($entityType, static::ALLOWED_ENTITY_TYPES)) {
      $elements = [];
      $linkTo = $this->getSetting('link_to');
      $displayInfo = $this->getSetting('info_shown');

      foreach ($this->getEntitiesToView($items, $langcode) as $delta => $entity) {
        $itemUrl = NULL;
        $label = $this->getSetting('link_text') ?: $entity->label();

        if ($fileHelper = FileHelper::fromEntity($entity)) {
          $info = [];

          if (!empty($displayInfo['mime'])) {
            $filenameParts = explode('.', $fileHelper->getFilename());
            $ext = end($filenameParts);

            $info[] = $this->mimeManager->getDisplayName($ext) ?: $fileHelper->getMime();
          }
          if (!empty($displayInfo['size'])) {
            $info[] = $this->readableFileSize($fileHelper->getSize());
          }

          $label .= (empty($info) ? '' : ' (' . implode(', ', $info) . ')');
        }

        if (!empty($linkTo)) {
          $itemUrl = $fileHelper && ($linkTo === 'file' || $entity instanceof File)
            ? Url::fromUri($fileHelper->buildRawUrl())
            : $entity->toUrl('canonical');
        }

        if (!empty($itemUrl)) {
          $elements[$delta] = [
            '#type' => 'link',
            '#title' => $label,
            '#url' => $itemUrl,
          ];
        }
        else {
          $elements[$delta] = [
            '#prefix' => '<span>',
            '#suffix' => '</span>',
            '#plain_text' => $label,
          ];
        }
      }

      return $elements;
    }

    // Unrecognized entity type, so we return an empty render array.
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'link_to' => 'file',
      'link_text' => NULL,
      'info_shown' => [
        'mime' => 'mime',
        'size' => 'size',
      ],
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = [];

    $infoDisplay = array_filter($this->getSetting('info_shown'));
    if (!empty($infoDisplay)) {
      $summary[] = $this->t('Displaying file info: %display', [
        '%display' => implode(', ', $infoDisplay),
      ]);
    }

    $linkTo = $this->getSetting('link_to');
    $linkOpts = $this->getLinkOptions();

    if (!empty($linkTo) && isset($linkOpts[$linkTo])) {
      $summary[] = $this->t('Linked to <strong>@target</strong>', [
        '@target' => $linkOpts[$linkTo],
      ]);
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $form = parent::settingsForm($form, $form_state);

    $form['link_to'] = [
      '#type' => 'select',
      '#title' => $this->t('Link to:'),
      '#options' => ['' => $this->t('No link')] + $this->getLinkOptions(),
      '#default_value' => $this->getSetting('link_to'),
    ];

    $form['link_text'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Link text:'),
      '#default_value' => $this->getSetting('link_text'),
      '#description' => $this->t('Leave blank to use default'),
    ];

    $form['info_shown'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Info Shown'),
      '#description' => $this->t('Use the options to determine what will be displayed'),
      '#options' => [
        'mime' => $this->t('File mime'),
        'size' => $this->t('File size'),
      ],
      '#default_value' => $this->getSetting('info_shown'),
    ];

    return $form;
  }

}
