<?php

namespace Drupal\toolshed_media\Routing;

use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\Session\AccountInterface;
use Drupal\toolshed_media\Utility\FileHelper;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\HttpKernel\Event\RequestEvent;

/**
 * Event subscriber handler which redirects Media requests to file.
 *
 * By default the media entity canonical display page, points to fully rendered
 * display page (entity display mode), but for most non-administrators, we
 * tend want to redirect to the file directly.
 */
class MediaRedirectRequestSubscriber implements EventSubscriberInterface {

  /**
   * Route used to determine how the request needs to be handled.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $routeMatch;

  /**
   * The account to use when determining if the request should redirect.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $account;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Generate a new MediaRedirectRequestSubscriber event listener.
   *
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The routing match interface to determine what route the request is
   *   acting on, and if it matches a media entity request.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   An account / session proxy representing the current user to evaluate
   *   the permissions and handling of the account retrieval.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(RouteMatchInterface $route_match, AccountInterface $account, EntityTypeManagerInterface $entity_type_manager) {
    $this->routeMatch = $route_match;
    $this->account = $account;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return [
      KernelEvents::REQUEST => 'onHandleRequest',
    ];
  }

  /**
   * Respond to the Request, and redirect media pages if appropriate.
   *
   * Determine if the request is for the canonical media display page and
   * decide if the page needs to get redirected to the media source file
   * if the user access and media type settings call for it.
   *
   * @param \Symfony\Component\HttpKernel\Event\RequestEvent $event
   *   The event information. This will contain the current kernel object and
   *   information about the request, since the event is responding to a
   *   KernelEvents::REQUEST event.
   */
  public function onHandleRequest(RequestEvent $event) {
    // Ensure that the current route is for a media entity display, and
    // get the media entity that is being requested.
    $media = $this->routeMatch->getRouteName() === 'entity.media.canonical'
      ? $this->routeMatch->getParameter('media')
      : NULL;

    if ($media && !$media->access('update', $this->account)) {
      /** @var \Drupal\media\MediaTypeInterface $mediaType */
      $mediaType = $media->bundle->entity;

      if (!$mediaType) {
        return;
      }

      if (!empty($mediaType->getThirdPartySetting('toolshed_media', 'redirect_to_file'))) {
        if ($fileHelper = FileHelper::fromEntity($media)) {
          $response = new TrustedRedirectResponse($fileHelper->buildRawUrl(FALSE));

          $cacheMetadata = new CacheableMetadata();
          $cacheMetadata->addCacheableDependency($mediaType);
          $cacheMetadata->addCacheableDependency($media);
          $cacheMetadata->setCacheContexts(['user.permissions']);
          $response->addCacheableDependency($cacheMetadata);

          $event->setResponse($response);
        }
        else {
          // Media referenced file is missing, return a 404.
          throw new NotFoundHttpException();
        }
      }
    }
  }

}
