<?php

namespace Drupal\toolshed_search\Plugin\Block;

use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Block\BlockBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\FormState;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Session\AccountInterface;
use Drupal\views\ViewExecutableFactory;
use Drupal\toolshed_search\Form\ToolshedSearchFilterForm;
use Drupal\toolshed_search\Plugin\ToolshedSearchAutocompleteInterface;
use Drupal\toolshed_search\Plugin\ToolshedSearchAutocompletePluginManager;
use Drupal\views\ViewExecutable;

/**
 * Block for creating configurable exposed search blocks.
 *
 * @Block(
 *   id = "toolshed_search_form",
 *   admin_label = @Translation("Toolshed: Search block"),
 *   category = @Translation("Search API"),
 * )
 */
class ToolshedSearchBlock extends BlockBase implements ContainerFactoryPluginInterface {

  /**
   * The loaded view or FALSE if the view is no longer available.
   *
   * @var \Drupal\views\ViewExecutable|false
   */
  protected $loadedView;

  /**
   * An array of configured views filters to exposed in the block form.
   *
   * @var \Drupal\views\Plugin\views\filter\FilterPluginBase[]
   */
  protected $filters;

  /**
   * Drupal form builder interface for generating form structures.
   *
   * @var \Drupal\Core\Form\FormBuilderInterface
   */
  protected $formBuilder;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Creates executatable views from views entity config instances.
   *
   * @var \Drupal\views\ViewExecutableFactory
   */
  protected $viewFactory;

  /**
   * The search autocomplete plugin manager.
   *
   * @var \Drupal\toolshed_search\Plugin\ToolshedSearchAutocompletePluginManager
   */
  protected $autocompleteManager;

  /**
   * Create a new instance of the toolshed search block.
   *
   * @param array $configuration
   *   The plugin configurations for the block setup.
   * @param string $plugin_id
   *   The unique ID of the plugin.
   * @param mixed $plugin_definition
   *   The plugin definition from the plugin discovery.
   * @param \Drupal\Core\Form\FormBuilderInterface $form_builder
   *   The Drupal form builder service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\views\ViewExecutableFactory $view_executable_factory
   *   Creates executatable views from views entity config instances.
   * @param \Drupal\toolshed_search\Plugin\ToolshedSearchAutocompletePluginManager $autocomplete_manager
   *   The search autocomplete plugin manager.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, FormBuilderInterface $form_builder, EntityTypeManagerInterface $entity_type_manager, ViewExecutableFactory $view_executable_factory, ToolshedSearchAutocompletePluginManager $autocomplete_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->formBuilder = $form_builder;
    $this->entityTypeManager = $entity_type_manager;
    $this->viewFactory = $view_executable_factory;
    $this->autocompleteManager = $autocomplete_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('form_builder'),
      $container->get('entity_type.manager'),
      $container->get('views.executable'),
      $container->get('plugin.manager.toolshed_search_autocomplete')
    );
  }

  /**
   * Retrieve a view and display settings based a view and display ID.
   *
   * @return \Drupal\views\ViewExecutable|null
   *   The view entity loaded, with the proper display configured. Will return
   *   NULL if the view or the display are no longer available.
   */
  protected function loadView($viewName): ?ViewExecutable {
    [$viewId, $display] = explode(':', $viewName);

    if ($viewId && $display) {
      $entity = $this->entityTypeManager
        ->getStorage('view')
        ->load($viewId);

      if ($entity) {
        $view = $this->viewFactory->get($entity);
        return $view->setDisplay($display) ? $view : NULL;
      }
    }

    return NULL;
  }

  /**
   * Retrieve the view configured to work with this block.
   *
   * @return \Drupal\views\ViewExecutable|false
   *   The view entity loaded, with the proper display configured. Will return
   *   NULL if the view or the display are no longer available.
   */
  public function getView() {
    if (!isset($this->loadedView)) {
      $this->loadedView = $this->loadView($this->configuration['view']) ?: FALSE;
    }

    return $this->loadedView;
  }

  /**
   * Get the views exposed filter plugins configured to be rendred.
   *
   * @param bool $reset
   *   If true the method will always generate the list of exposed filters.
   *
   * @return \Drupal\views\Plugin\views\filter\FilterPluginBase[]
   *   List of plugin filters which are exposed for the view and configured to
   *   be built by the block configurations.
   */
  public function getFilters($reset = FALSE) {
    if (!isset($this->filters) || $reset) {
      $this->filters = [];

      if ($view = $this->getView()) {
        $config = $this->getConfiguration();
        $filters = $view->display_handler->getHandlers('filter');

        /** @var \Drupal\views\Plugin\views\filter\FilterPluginBase $filter */
        foreach (array_intersect_key($filters, $config['filters']) as $filterId => $filter) {
          if ($filter->canExpose() && $filter->isExposed()) {
            $this->filters[$filterId] = $filter;
          }
        }
      }
    }

    return $this->filters;
  }

  /**
   * Gets a loaded instance of the search autocomplete plugin for this block.
   *
   * @return \Drupal\toolshed_search\Plugin\ToolshedSearchAutocompleteInterface|null
   *   The search autocomplete plugin if one was configured for this block.
   */
  public function getAutocomplete(): ?ToolshedSearchAutocompleteInterface {
    try {
      if (!empty($this->configuration['autocomplete']['plugin'])) {
        ['plugin' => $pluginId, 'config' => $config] = $this->configuration['autocomplete'] + [
          'config' => [],
        ];

        return $this->autocompleteManager->createInstance($pluginId, $config);
      }
    }
    catch (PluginNotFoundException) {
      // Plugin missing, just treat this as no autocomplete available.
    }

    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function blockAccess(AccountInterface $account) {
    if ($view = $this->getView()) {
      $access = $view->access($view->current_display, $account);
      return AccessResult::allowedIf($access);
    }

    return AccessResult::forbidden();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTags() {
    $tags = parent::getCacheTags();

    if ($view = $this->getView()) {
      $viewTags = $view->display_handler
        ->getCacheMetadata()
        ->getCacheTags();

      $tags = Cache::mergeTags($viewTags, $tags);
    }

    return $tags;
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheContexts() {
    $contexts = parent::getCacheContexts();

    if ($view = $this->getView()) {
      $viewContexts = $view->display_handler
        ->getCacheMetadata()
        ->getCacheContexts();

      $contexts = Cache::mergeContexts($viewContexts, $contexts);

      // Add cache contexts from each of the configured views filters.
      foreach ($this->getFilters() as $filter) {
        $contexts = Cache::mergeContexts($filter->getCacheContexts(), $contexts);
      }
    }

    return $contexts;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'view' => ':',
      'autocomplete' => '',
      'filters' => [],
      'action' => 'view_url',
      'submit_label' => $this->t('Search'),
      'submit_class_names' => [],
      'variant' => NULL,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function build() {
    if (!empty($this->configuration['filters']) && $this->getView()) {
      $state = new FormState();
      $state->setMethod('GET');
      $state->setRequestMethod('GET');
      $state->disableCache();
      $state->disableRedirect();
      $state->addBuildInfo('args', [$this]);

      return $this->formBuilder->buildForm(ToolshedSearchFilterForm::class, $state);
    }

    return [];
  }

  /**
   * Get the available views which can be selected to create filter blocks for.
   *
   * @return array
   *   Array of views that can be the target of this search filters form.
   */
  protected function getViewOptions() {
    /** @var \Drupal\views\ViewEntityInterface $views */
    $views = $this->entityTypeManager
      ->getStorage('view')
      ->loadByProperties(['status' => TRUE]);

    $options = ['' => '-- Select a View --'];
    foreach ($views as $viewId => $view) {
      if ($view->get('base_field') !== 'search_api_id' && !preg_match('/^search_api_index/', $view->get('base_table'))) {
        continue;
      }

      // Only create a blocks that have search API page displays to redirect to.
      foreach ($view->get('display') as $displayId => $display) {
        if ($display['display_plugin'] === 'page') {
          $options["$viewId:$displayId"] = $view->label() . " ($display[display_title])";
        }
      }
    }

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function blockForm($form, FormStateInterface $form_state) {
    $config = $this->getConfiguration();
    $parentState = $form_state instanceof SubformState ? $form_state->getCompleteFormState() : $form_state;

    $viewId = $parentState->getValue(['settings', 'view']);
    $viewId = !isset($viewId) ? $config['view'] : $viewId;
    $viewOptions = $this->getViewOptions();

    $form['view'] = [
      '#type' => 'select',
      '#title' => $this->t('View to run search'),
      '#required' => TRUE,
      '#options' => $viewOptions,
      '#default_value' => $viewId,
      '#description' => $this->t('Only search_api views with a page display appear in this list.'),
      '#ajax' => [
        'wrapper' => 'views-display-ajax-wrapper',
        'callback' => static::class . '::ajaxUpdateViewDisplays',
      ],
    ];

    $form['filters'] = [
      '#prefix' => '<div id="views-display-ajax-wrapper">',
      '#suffix' => '</div>',
    ];

    if (isset($viewOptions[$viewId]) && ($view = $this->loadView($viewId))) {
      $filterOpts = [];

      /** @var \Drupal\views\Plugin\views\filter\FilterPluginBase $handler */
      foreach ($view->display_handler->getHandlers('filter') as $fieldName => $handler) {
        if ($handler->isExposed()) {
          $filterOpts[$fieldName] = $handler->pluginTitle();
        }
      }

      $form['filters'] = [
        '#type' => 'checkboxes',
        '#title' => $this->t('Exposed filters'),
        '#required' => TRUE,
        '#options' => $filterOpts,
        '#default_value' => $this->configuration['filters'],
        '#description' => $this->t('Select the filter which should be used for the keyword input. Only exposed fulltext filters will appear in this list.'),
      ];

      $acOptions = [];
      foreach ($this->autocompleteManager->getDefinitions() as $pluginId => $value) {
        $autocomplete = $this->autocompleteManager->createInstance($pluginId, []);
        if ($autocomplete->isApplicable($view)) {
          $acOptions[$pluginId] = $value['label'];
        }
      }

      if ($acOptions) {
        $form['autocomplete'] = [
          '#type' => 'fieldset',
          '#title' => $this->t('Fulltext Autocomplete'),
          '#tree' => TRUE,

          'plugin' => [
            '#type' => 'select',
            '#title' => $this->t('Provider'),
            '#options' => ['' => $this->t('- None -')] + $acOptions,
            '#default_value' => $this->configuration['autocomplete']['plugin'] ?? '',
            '#description' => $this->t('Applies the search autocomplete to all the search_api fulltext filters.'),
          ],
          'config' => [
            '#type' => 'value',
            '#value' => [],
          ],
        ];
      }
      else {
        $form['autocomplete'] = [
          '#type' => 'value',
          '#value' => [
            'plugin' => '',
            'config' => [],
          ],
        ];
      }
    }

    $form['submit_label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Submit button text'),
      '#default_value' => $config['submit_label'],
    ];

    $form['submit_class_names'] = [
      '#type' => 'css_class',
      '#title' => $this->t('Button class names'),
      '#default_value' => $config['submit_class_names'],
      '#description' => $this->t('Additional CSS class names to apply to the submit button.'),
    ];

    $form['variant'] = [
      '#type' => 'textfield',
      '#title' => $this->t("Block variant"),
      '#pattern' => '^[\w_ ]*$',
      '#default_value' => $this->configuration['variant'],
      '#description' => $this->t('Optional extra label to provide info to the ARIA descriptions to provide context if the search block appears more than once on a page. (i.e. "footer" or "header")'),
    ];

    $form['action'] = [
      '#type' => 'select',
      '#title' => $this->t('Form submit target'),
      '#options' => [
        'view_url' => $this->t('The view display page'),
        'current_page' => $this->t('Submit to same page'),
      ],
      '#default_value' => $config['action'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function blockSubmit($form, FormStateInterface $form_state) {
    $values = $form_state->getValues();

    $this->configuration['view'] = $values['view'] ?? ':';
    $this->configuration['filters'] = array_filter($values['filters']);
    $this->configuration['autocomplete'] = $values['autocomplete'];
    $this->configuration['action'] = $values['action'];
    $this->configuration['submit_label'] = trim($values['submit_label']);
    $this->configuration['submit_class_names'] = $form_state->getValue('submit_class_names');
    $this->configuration['variant'] = trim($values['variant']);
  }

  /**
   * Update the block configuration form after the view has been selected.
   *
   * @param array $form
   *   Structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form and build information.
   *
   * @return array
   *   Renderable array of content to be replaced using AJAX.
   */
  public static function ajaxUpdateViewDisplays(array $form, FormStateInterface $form_state) {
    $trigger = $form_state->getTriggeringElement();
    $parents = array_slice($trigger['#array_parents'], 0, -1);
    $parents[] = 'filters';

    return NestedArray::getValue($form, $parents);
  }

}
